package LOGCHECK;

# +-------------------------------------------------------------------+
# |            S E N D M A I L   L O G C H E C K   (v1.2)             |
# |                                                                   |
# | Copyright Gerd Tentler                www.gerd-tentler.de/tools   |
# | Created: Mar. 29, 2005                Last modified: Feb. 3, 2006 |
# +-------------------------------------------------------------------+
# | This program may be used and hosted free of charge by anyone for  |
# | personal purpose as long as this copyright notice remains intact. |
# |                                                                   |
# | Obtain permission before selling the code for this program or     |
# | hosting this software on a commercial website or redistributing   |
# | this software over the Internet or in any other medium. In all    |
# | cases copyright must remain intact.                               |
# +-------------------------------------------------------------------+
#
#======================================================================================================
# Example:
#
#   use LOGCHECK;
#
#   $log = LOGCHECK->new("/path/to/sendmail.log");
#   print $log->report();
#
# Returns HTML or ASCII report
#======================================================================================================

#------------------------------------------------------------------------------------------------------
# Subs
#------------------------------------------------------------------------------------------------------

# CONSTRUCTOR
sub new {
  my $logfile = $_[1];
  my $self = {};
  bless $self;
  $self->initialize();
  if($logfile) { $self->{maillog} = $logfile; }
  return $self;
}

sub initialize {
#----------------------------------------------------------------------------------------------------
# Configuration
#----------------------------------------------------------------------------------------------------
  my $this = shift;
  $this->{maillog}        = "/var/log/mail";      # path to sendmail log
  $this->{format}         = "HTML";               # format of report ("HTML" or "ASCII")
  $this->{entries}        = 10;                   # number of entries in top ranking

  # only HTML format:
  $this->{titleFont}      = "Arial, Helvetica";   # title font family (CSS spec)
  $this->{titleFontSize}  = 16;                   # title font size (pixels)
  $this->{titleFontColor} = "#000000";            # title font color
  $this->{titleBGColor}   = "#C0E0FF";            # title background color
  $this->{titleBorder}    = "2px groove white";   # title border (CSS spec)

  $this->{headFont}       = "Arial, Helvetica";   # table head font family (CSS spec)
  $this->{headFontSize}   = 12;                   # table head font size (pixels)
  $this->{headFontColor}  = "#FFFFFF";            # table head font color
  $this->{headBGColor}    = "#A0C0E0";            # table head background color
  $this->{headBorder}     = "2px groove white";   # table head border (CSS spec)

  $this->{cellFont}       = "Arial, Helvetica";   # table cell font family (CSS spec)
  $this->{cellFontSize}   = 12;                   # table cell font size (pixels)
  $this->{cellFontColor}  = "#000000";            # table cell font color
  $this->{cellBGColor}    = "#E0F0FF";            # table cell background color
  $this->{cellBorder}     = "2px groove white";   # table cell border (CSS spec)

  $this->{graphFont}      = "Arial, Helvetica";   # graph font family (CSS spec)
  $this->{graphFontSize}  = 12;                   # graph font size (pixels)
  $this->{graphFontColor} = "#000000";            # graph font color
  $this->{graphBGColor}   = "#E0F0FF";            # graph background color
  $this->{graphBorder}    = "2px groove white";   # graph border (CSS spec)

  $this->{barWidth}       = 10;                   # graph bar width (pixels)
  $this->{barBorder}      = "2px outset white";   # graph bar border (CSS spec)
  $this->{barColor1}      = "#0000FF";            # graph bar color 1
  $this->{barColor2}      = "#FF0000";            # graph bar color 2

  $this->{labelFont}      = "Arial, Helvetica";   # graph label font family (CSS spec)
  $this->{labelFontSize}  = 12;                   # graph label font size (pixels)
  $this->{labelFontColor} = "#FFFFFF";            # graph label font color
  $this->{labelBGColor}   = "#A0C0E0";            # graph label background color
  $this->{labelBorder}    = "2px groove white";   # graph label border (CSS spec)

  # don't change!
  $this->{filter} = ();
  $this->{line1}  = '=============================================================================================';
  $this->{line2}  = '---------------------------------------------------------------------------------------------';
}

sub set_style {
  my $this = shift;
  my $style = '<style> .clsTitle { ';
  $style .= 'font-family: ' . $this->{titleFont} . '; ';
  $style .= 'font-size: ' . $this->{titleFontSize} . 'px; ';
  $style .= 'font-weight: bold; ';
  $style .= 'color: ' . $this->{titleFontColor} . '; ';
  if($this->{titleBGColor}) { $style .= 'background-color: ' . $this->{titleBGColor} . '; '; }
  if($this->{titleBorder}) { $style .= 'border: ' . $this->{titleBorder} . '; '; }
  $style .= '} .clsTableHead { ';
  $style .= 'font-family: ' . $this->{headFont} . '; ';
  $style .= 'font-size: ' . $this->{headFontSize} . 'px; ';
  $style .= 'font-weight: bold; ';
  $style .= 'color: ' . $this->{headFontColor} . '; ';
  if($this->{headBGColor}) { $style .= 'background-color: ' . $this->{headBGColor} . '; '; }
  if($this->{headBorder}) { $style .= 'border: ' . $this->{headBorder} . '; '; }
  $style .= '} .clsTableCell { ';
  $style .= 'font-family: ' . $this->{cellFont} . '; ';
  $style .= 'font-size: ' . $this->{cellFontSize} . 'px; ';
  $style .= 'color: ' . $this->{cellFontColor} . '; ';
  if($this->{cellBGColor}) { $style .= 'background-color: ' . $this->{cellBGColor} . '; '; }
  if($this->{cellBorder}) { $style .= 'border: ' . $this->{cellBorder} . '; '; }
  $style .= '} .clsGraph1 { ';
  if($this->{graphBGColor}) { $style .= 'background-color: ' . $this->{graphBGColor} . '; '; }
  if($this->{graphBorder}) { $style .= 'border: ' . $this->{graphBorder} . '; '; }
  $style .= '} .clsGraph2 { ';
  $style .= 'font-family: ' . $this->{graphFont} . '; ';
  $style .= 'font-size: ' . $this->{graphFontSize} . 'px; ';
  $style .= 'color: ' . $this->{graphFontColor} . '; ';
  $style .= '} .clsBar1 { ';
  $style .= 'background-color: ' . $this->{barColor1} . '; ';
  if($this->{barBorder}) { $style .= 'border: ' . $this->{barBorder} . '; '; }
  $style .= '} .clsBar2 { ';
  $style .= 'background-color: ' . $this->{barColor2} . '; ';
  if($this->{barBorder}) { $style .= 'border: ' . $this->{barBorder} . '; '; }
  $style .= '} .clsLabel { ';
  $style .= 'font-family: ' . $this->{labelFont} . '; ';
  $style .= 'font-size: ' . $this->{labelFontSize} . 'px; ';
  $style .= 'font-weight: bold; ';
  $style .= 'text-align: center; ';
  $style .= 'color: ' . $this->{labelFontColor} . '; ';
  if($this->{labelBGColor}) { $style .= 'background-color: ' . $this->{labelBGColor} . '; '; }
  if($this->{labelBorder}) { $style .= 'border: ' . $this->{labelBorder} . '; '; }
  $style .= "} </style>\n";
  return $style;
}

sub get_entries {
  my $this = shift;
  my %log = ();

  if(open(LOG, '<' . $this->{maillog})) {
    my @mID = ();
    my ($row, $id, $time, $dir, $email, $ctladdr);

    while($row = <LOG>) {

      if($row =~ m/ sendmail.+: ([a-z0-9]+): /io) {
        $id = $1;

        if($row =~ m/^([a-z]+ +[0-9]+ [0-9:]+) .+ (from|to)=<?([^<>, ]+)>?(, ctladdr=<?([^@<>, ]+@[^<>, ]+)>?)?/io) {
          $time = _trim($1);
          $dir = $2;
          $email = lc(_trim($3));
          $ctladdr = lc(_trim($5));

          if(!($email =~ m/@/)) { $email = $ctladdr; }

          if(!$log{$id}) { $log{$id} = (); }

          if($email ne $log{$id}{from} || !$log{$id}{stat}) {
            $log{$id}{time} = $time;
            $log{$id}{$dir} = $email;

            if($row =~ m/ size=([0-9]+)/io) { $log{$id}{size} = $1; }
            if($row =~ m/ stat=(.+)/io) { $log{$id}{stat} = lc(_trim($1)); }
          }
        }
      }
    }
    close(LOG);

    if(keys(%log)) {

      foreach $id (keys(%log)) {

        if($log{$id}{stat} =~ m/sent \(([a-z0-9]+) message/io) {
          if($log{$1}) { delete $log{$1}; }
        }

        if(!$log{$id}{size} || !$log{$id}{from} || !$log{$id}{to}) { delete $log{$id}; }
        else {
          if($this->{filter}{from} && !($log{$id}{from} =~ m/$this->{filter}{from}/io)) { delete $log{$id}; }
          if($this->{filter}{to} && !($log{$id}{to} =~ m/$this->{filter}{to}/io)) { delete $log{$id}; }
          if($this->{filter}{stat} && !($log{$id}{stat} =~ m/$this->{filter}{stat}/io)) { delete $log{$id}; }
        }
      }
    }
  }
  return %log;
}

sub draw_bar {
  my ($this, $class, $width, $height, $val) = @_;
  my $bar  = '<table border=0 cellspacing=0 cellpadding=0><tr>';
  $bar .= '<td class="' . $class . '"><table border=0 cellspacing=0 cellpadding=0><tr>';
  $bar .= '<td width=' . $width . ' height=' . $height . '></td>';
  $bar .= '</tr></table></td>';
  if($val) { $bar .= '<td class="clsGraph2">&nbsp;' . $val . '</td>'; }
  $bar .= '</tr></table>';
  return $bar;
}

sub set_status {
  my ($this, $stat) = @_;
  if($stat =~ m/^timeout/io) { $stat = 'timeout'; }
  else {
    $stat =~ s/([a-z0-9' ]+).*/$1/io;
    $stat = _trim($stat);
  }
  return $stat;
}

sub build_graph {
  my ($this, $log) = @_;
  my (%cnt, %size) = ();
  my ($sum_cnt, $sum_size, $max_cnt, $max_size) = 0;
  my ($graph, $id, $percent1, $percent2, $length1, $length2, $val1, $val2, $stat, $i);

  foreach $id (keys(%{$log})) {
    $stat = $this->set_status($log->{$id}{stat});
    $sum_cnt++;
    $cnt{$stat}++;
    if($cnt{$stat} > $max_cnt) { $max_cnt = $cnt{$stat}; }

    $sum_size += $log->{$id}{size};
    $size{$stat} += $log->{$id}{size};
    if($size{$stat} > $max_size) { $max_size = $size{$stat}; }
  }
  my %topcnt = _ksort(\%cnt);
  my %topsize = _ksort(\%size);

  if($this->{format} eq 'HTML') {
    $graph  = "<table border=0 cellspacing=2 cellpadding=2 width=100%><tr align=center>\n";
    $graph .= '<td class="clsTitle" colspan=2>';
    $graph .= '<table border=0 cellspacing=0 cellpadding=4><tr>';
    $graph .= '<td>' . $this->draw_bar('clsBar1', 12, 12) . '</td>';
    $graph .= '<td class="clsGraph2"><b>Messages: ' . $sum_cnt . '</b></td>';
    $graph .= '<td width=20>&nbsp;</td>';
    $graph .= '<td>' . $this->draw_bar('clsBar2', 12, 12) . '</td>';
    $graph .= '<td class="clsGraph2"><b>Traffic: ' . _number_format($sum_size / 1024, 2) . ' KB</b></td>';
    $graph .= "</tr></table></td></tr>\n";

    for($i = 0; $i < keys(%topcnt); $i++) {
      $percent1 = _number_format($topcnt{$i}{val} * 100 / $sum_cnt, 1);
      $percent2 = _number_format($topsize{$i}{val} * 100 / $sum_size, 1);
      $length1 = _ceil($topcnt{$i}{val} * 100 / $max_cnt);
      $length2 = _ceil($topsize{$i}{val} * 100 / $max_size);
      $val1 = $topcnt{$i}{val} . " ($percent1%)";
      $val2 = _number_format($topsize{$i}{val} / 1024, 2) . " KB ($percent2%)";

      $graph .= '<tr>';
      $graph .= '<td class="clsLabel">' . $topcnt{$i}{key} . '</td>';
      $graph .= '<td class="clsGraph1"><table border=0 cellspacing=0 cellpadding=0><tr>';
      $graph .= '<td>' . $this->draw_bar('clsBar1', $length1*3, $this->{barWidth}, $val1) . '</td>';
      $graph .= "</tr><tr><td height=2></td></tr><tr>\n";
      $graph .= '<td>' . $this->draw_bar('clsBar2', $length2*3, $this->{barWidth}, $val2) . '</td>';
      $graph .= '</tr></table></td>';
      $graph .= "</tr>\n";
    }
    $graph .= "</table>\n";
  }
  else {
    my $bar1 = 'ooooooooooooooooooooooooooooooooooooooooooooo';
    my $bar2 = 'OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO';
    my $line = $this->{line2};

    $graph = "oo Messages: $sum_cnt\nOO Traffic : " . _number_format($sum_size / 1024, 2) . " KB\n";

    for($i = 0; $i < keys(%topcnt); $i++) {
      $percent1 = _number_format($topcnt{$i}{val} * 100 / $sum_cnt, 1);
      $percent2 = _number_format($topsize{$i}{val} * 100 / $sum_size, 1);
      $length1 = _ceil($topcnt{$i}{val} * 45 / $max_cnt);
      $length2 = _ceil($topsize{$i}{val} * 45 / $max_size);
      $val1 = $topcnt{$i}{val} . " ($percent1%)";
      $val2 = _number_format($topsize{$i}{val} / 1024, 2) . " KB ($percent2%)";

      $graph .= "$line\n";
      $graph .= sprintf('%-20s: %-45s %s', $topcnt{$i}{key}, substr($bar1, 0, $length1), $val1) . "\n";
      $graph .= sprintf('%-20s: %-45s %s', '', substr($bar2, 0, $length2), $val2) . "\n";
    }
  }
  return $graph;
}

sub build_table_messages {
  my ($this, $log) = @_;
  my %top = ();
  my ($table, $id, $i, $j, $from, $to);

  foreach $id (keys(%{$log})) {
    $i = 0;

    while($log->{$id}{size} <= $top{$i}{size} && $i < $this->{entries}) { $i++; }

    if($i < $this->{entries}) {
      for($j = $this->{entries}-1; $j > $i; $j--) { $top{$j} = $top{$j-1}; }
      $top{$i} = $log->{$id};
      $top{$i}{stat} = $this->set_status($top{$i}{stat});
    }
  }

  if($this->{format} eq 'HTML') {
    $table  = '<table border=0 cellspacing=2 cellpadding=2 width=100%><tr align=center>';
    $table .= '<td colspan=6 class="clsTitle">Messages: Top ' . $this->{entries} . ' by size</td>';
    $table .= '</tr><tr align=center>';
    $table .= '<td class="clsTableHead">Rank</td>';
    $table .= '<td class="clsTableHead">Time</td>';
    $table .= '<td class="clsTableHead">From</td>';
    $table .= '<td class="clsTableHead">To</td>';
    $table .= '<td class="clsTableHead">Size</td>';
    $table .= '<td class="clsTableHead">Status</td>';
    $table .= "</tr>\n";

    for($i = 0; $i < keys(%top); $i++) {
      if($top{$i}{size}) {
        $table .= '<tr>';
        $table .= '<td class="clsTableCell" align=center><b>' . ($i+1) . '</b></td>';
        $table .= '<td class="clsTableCell" nowrap>' . $top{$i}{time} . '</td>';
        $table .= '<td class="clsTableCell" nowrap>' . $top{$i}{from} . '</td>';
        $table .= '<td class="clsTableCell" nowrap>' . $top{$i}{to} . '</td>';
        $table .= '<td class="clsTableCell" align=right>' . $top{$i}{size} . '</td>';
        $table .= '<td class="clsTableCell">' . $top{$i}{stat} . '</td>';
        $table .= "</tr>\n";
      }
    }
    $table .= "</tr></table>\n";
  }
  else {
    my $line = $this->{line2};
    $table  = 'Messages: Top ' . $this->{entries} . " by size\n$line\n";
    $table .= sprintf('%4s  %-15s  %-22s  %-22s  %8s  %s', '', 'Time', 'From', 'To', 'Size', 'Status') . "\n$line\n";

    for($i = 0; $i < keys(%top); $i++) {
      if($top{$i}{size}) {
        $from = $top{$i}{from};
        $to = $top{$i}{to};

        if(length($from) > 22) { $from = substr($from, 0, 20) . '..'; }
        if(length($to) > 22) { $to = substr($to, 0, 20) . '..'; }

        $table .= sprintf('%4d  %-15s  %-22s  %-22s  %8d  %s', ($i+1), $top{$i}{time}, $from,
                          $to, $top{$i}{size}, $top{$i}{stat}) . "\n";
      }
    }
  }
  return $table;
}

sub build_table_addresses {
  my ($this, $log, $type) = @_;
  my (%cnt, %size) = ();
  my $title = ($type eq 'from') ? 'Senders' : 'Recipients';
  my ($table, $id, $email, $i);

  foreach $id (keys(%{$log})) {

    if($log->{$id}{size}) {
      $email = $log->{$id}{$type};
      if(length($email) > 65) { $email = substr($email, 0, 63) . '..'; }
      $cnt{$email}++;
      $size{$email} += $log->{$id}{size};
    }
  }
  my %top = _arsort(\%size, $this->{entries});

  if($this->{format} eq 'HTML') {
    $table  = '<table border=0 cellspacing=2 cellpadding=2 width=100%><tr align=center>';
    $table .= '<td colspan=6 class="clsTitle">' . $title . ': Top ' . $this->{entries} . ' by size</td>';
    $table .= '</tr><tr align=center>';
    $table .= '<td class="clsTableHead" width=8%>Rank</td>';
    $table .= '<td class="clsTableHead" width=62%>Address</td>';
    $table .= '<td class="clsTableHead" width=15%>Messages</td>';
    $table .= '<td class="clsTableHead" width=15%>Size</td>';
    $table .= "</tr>\n";

    for($i = 0; $i < keys(%top) && $i < $this->{entries}; $i++) {
      $table .= '<tr>';
      $table .= '<td class="clsTableCell" align=center width=8%><b>' . ($i+1) . '</b></td>';
      $table .= '<td class="clsTableCell" nowrap width=62%>' . $top{$i}{key} . '</td>';
      $table .= '<td class="clsTableCell" align=center width=15%>' . $cnt{$top{$i}{key}} . '</td>';
      $table .= '<td class="clsTableCell" align=right width=15%>' . $top{$i}{val} . '</td>';
      $table .= "</tr>\n";
    }
    $table .= "</table>\n";
  }
  else {
    my $line = $this->{line2};
    $table  = $title . ': Top ' . $this->{entries} . " by size\n$line\n";
    $table .= sprintf('%4s  %-65s  %8s  %8s', '', 'Address', 'Messages', 'Size') . "\n$line\n";

    for($i = 0; $i < keys(%top) && $i < $this->{entries}; $i++) {
      $email = $top{$i}{key};
      $table .= sprintf('%4d  %-65s  %8d  %8d', ($i+1), $email, $cnt{$email}, $top{$i}{val}) . "\n";
    }
  }
  return $table;
}

sub show_filter {
  my $this = shift;
  my $from = $this->{filter}{from};
  my $to = $this->{filter}{to};
  my $stat = $this->{filter}{stat};
  my $table;

  if($this->{format} eq 'HTML') {
    $table  = "<table border=0 cellspacing=2 cellpadding=2 width=100%><tr align=center>\n";
    $table .= '<td class="clsTitle" colspan=3>Filter</td>';
    $table .= "</tr><tr align=center>\n";
    $table .= '<td class="clsTableHead" width=35%>From</td>';
    $table .= '<td class="clsTableHead" width=35%>To</td>';
    $table .= '<td class="clsTableHead" width=30%>Status</td>';
    $table .= "</tr><tr align=center>\n";
    $table .= '<td class="clsTableCell">' . ($from ? $from : '&nbsp;') . '</td>';
    $table .= '<td class="clsTableCell">' . ($to ? $to : '&nbsp;') . '</td>';
    $table .= '<td class="clsTableCell">' . ($stat ? $stat : '&nbsp;') . '</td>';
    $table .= "</tr></table>\n";
  }
  else {
    my $line = $this->{line2};
    $table  = "FILTER\n$line\n";
    $table .= sprintf('%-35s  %-35s  %s', 'From', 'To', 'Status') . "\n$line\n";

    if(length($from) > 35) { $from = substr($from, 0, 33) . '..'; }
    if(length($to) > 35) { $to = substr($to, 0, 33) . '..'; }

    $table .= sprintf('%-35s  %-35s  %s', $from, $to, $stat) . "\n";
  }
  return $table;
}

sub report {
  my $this = shift;
  my %log = $this->get_entries();
  my ($content, $time);

  if($this->{format} eq 'HTML') {
    $content  = $this->set_style();
    $content .= '<table border=0 cellspacing=0 cellpadding=0><tr><td align=center>';
    $content .= '<font face="Arial, Helvetica" size=3><b>';
    $content .= 'S E N D M A I L &nbsp; L O G C H E C K<br>';
    $content .= '<font size=2>&copy; by Gerd Tentler</font>';
    $content .= "</b></font><br><br></td></tr><tr><td align=center>\n";

    if(keys(%{$this->{filter}})) { $content .= $this->show_filter() . "<br>\n"; }

    if(keys(%log)) {
      $content .= $this->build_graph(\%log) . "<br>\n";
      $content .= $this->build_table_messages(\%log) . "<br>\n";
      $content .= $this->build_table_addresses(\%log, 'from') . "<br>\n";
      $content .= $this->build_table_addresses(\%log, 'to') . "\n";
    }
    else { $content .= '<div class="clsTableCell">No entries found.</div>'; }

    $content .= '</td></tr></table>';
  }
  else {
    my $line = $this->{line1};
    $content = "S E N D M A I L   L O G C H E C K\n(c) by Gerd Tentler\n\n";

    if(keys(%{$this->{filter}})) { $content .= "$line\n" . $this->show_filter() . "$line\n\n"; }

    if(keys(%log)) {
      $content .= "$line\n" . $this->build_graph(\%log) . "$line\n\n";
      $content .= "$line\n" . $this->build_table_messages(\%log) . "$line\n\n";
      $content .= "$line\n" . $this->build_table_addresses(\%log, 'from') . "$line\n\n";
      $content .= "$line\n" . $this->build_table_addresses(\%log, 'to') . "$line\n\n";
    }
    else { $content .= "No entries found.\n\n"; }
  }
  return $content;
}

#------------------------------------------------------------------------------------------------------
# PRIVATE
#------------------------------------------------------------------------------------------------------

sub _trim {
  my $s = shift;
  $s =~ s/^\s+|\s+$//g;
  return $s;
}

sub _round {
  my $fl = shift;
  return ($fl >= 0) ? int($fl + 0.5) : int($fl - 0.5);
}

sub _ceil {
  my $fl = shift;
  return int($fl + 0.9);
}

sub _number_format {
  my ($percent, $percValuesDecimals) = @_;
  my $format;
  $format = ($percValuesDecimals > 0) ? 'f' : 'd';
  if($format eq 'd') { $percent = _round($percent); }
  return sprintf('%.' . $percValuesDecimals . $format, $percent);
}

sub _ksort {
  my $in = shift;
  my %top = ();
  my $i = 0;

  foreach $key (sort keys(%{$in})) {
    $top{$i}{key} = $key;
    $top{$i}{val} = $in->{$key};
    $i++;
  }
  return %top;
}

sub _arsort {
  my ($in, $max) = @_;
  my %top = ();
  my ($key, $i, $j, $start);

  foreach $key (keys(%{$in})) {
    $i = 0;

    while($in->{$key} <= $top{$i}{val} && ($i < $max || !$max)) { $i++; }

    if($i < $max || !$max) {
      for($j = keys(%top); $j > $i; $j--) {
        $top{$j}{key} = $top{$j-1}{key};
        $top{$j}{val} = $top{$j-1}{val};
      }
      $top{$i}{key} = $key;
      $top{$i}{val} = $in->{$key};
    }
    if(!$top{keys(%top)-1}{val}) { delete $top{keys(%top)-1}; }
  }
  return %top;
}

#------------------------------------------------------------------------------------------------------

return 1;
